import sys
import logging


def setup_logging(log, level=None):
    console_handler = logging.StreamHandler(sys.stderr)

    if level is not None:
        log.setLevel(level)
        console_handler.setLevel(level)

    console_handler.setFormatter(
        logging.Formatter("%(levelname)s: %(message)s"))

    log.addHandler(console_handler)


def setup_module_logging(name, level=None, *args, **kwargs):
    """ Sets up module-level logging
    """
    log = logging.getLogger(name, *args, **kwargs)
    setup_logging(log, level=level)
    return log



#!/bin/env python3

import os
import logging
import json

log = logging.getLogger(__name__)


class MigrationConfig:
    def __init__(self):
        config_file = 'config.json'
        if not os.path.exists(os.path.join('.', config_file)):
            raise FileNotFoundError
        with open(config_file, 'r') as outfile:
            data = json.load(outfile)
        log.debug('config: {}'.format(data))
        self.redmine_host = data['redmine']['host']
        self.redmine_project_url = self.redmine_host + '/' + data['redmine']['path']
        self.redmine_key = data['redmine']['key']
        self.gitlab_host = data['gitlab']['host']
        self.gitlab_project_url = self.gitlab_host + '/' + data['gitlab']['path']
        self.gitlab_key = data['gitlab']['key']
        self.cache_dir = os.path.join('.', 'redmine')
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)
            log.info('Create cache dir: {}'.format(self.cache_dir))



from itertools import chain
import os
import re
import logging
import json

from . import APIClient, Project

ANONYMOUS_USER_ID = 2

log = logging.getLogger(__name__)


class RedmineClient(APIClient):
    PAGE_MAX_SIZE = 100

    def get_auth_headers(self):
        return {"X-Redmine-API-Key": self.api_key}

    def get(self, *args, **kwargs):
        # In detail views, redmine encapsulate "foo" typed objects under a
        # "foo" key on the JSON.
        # noinspection PyCompatibility
        ret = super().get(*args, **kwargs)
        values = ret.values()
        if len(values) == 1:
            return list(values)[0]
        else:
            return ret

    def get_all_pages(self, *args, **kwargs):
        """ Iterates over API pagination for a given resource list
        """
        kwargs['params'] = kwargs.get('params', {})
        kwargs['params']['limit'] = self.PAGE_MAX_SIZE

        resp = self.get(*args, **kwargs)

        # Try to auto find the top-level key containing
        keys_candidates = (
            set(resp.keys()) - {'total_count', 'offset', 'limit'})

        assert len(keys_candidates) == 1
        res_list_key = list(keys_candidates)[0]

        result_pages = [resp[res_list_key]]
        if 'offset' not in resp:
            raise ValueError('HTTP response data is not paginated')

        while (resp['total_count'] - resp['offset'] - resp['limit']) > 0:
            kwargs['params']['offset'] = (kwargs['params'].get('offset', 0)
                                          + self.PAGE_MAX_SIZE)
            resp = self.get(*args, **kwargs)
            result_pages.append(resp[res_list_key])
        return chain.from_iterable(result_pages)


class RedmineProject(Project):
    REGEX_PROJECT_URL = re.compile(
        r'^(?P<base_url>https?://.*)/projects/(?P<project_name>[\w_-]+)$')

    REGEX_CATEGORY_PROJECT_URL = re.compile(
        r'^(?P<base_url>https?://.*)/project/(?P<category_name>[\w_-]+)/(?P<project_name>[\w_-]+)/?$')

    def __init__(self, url, *args, **kwargs):
        normalized_url = self._remove_category_in_url(url)
        # noinspection PyCompatibility
        super().__init__(normalized_url, *args, **kwargs)
        self.api_url = '{}.json'.format(self.public_url)
        self.instance_url = self._url_match.group('base_url')
        self.project = self.api.get(self.api_url)
        log.info('Got redmine project: {}'.format(self.get_id()))

    def get_id(self):
        return str(self.project.get('id', 0))

    def get_project(self):
        return self.project

    @classmethod
    def _remove_category_in_url(cls, url):
        """ If using categories, return the category-less URL

        eg:
          - category URL: https://example.com/project/dev/foobar/
          - category-less URL: https://example.com/projects/foobar/

        API endpoints are reachable only for category-less URLs.
        """
        m = cls.REGEX_CATEGORY_PROJECT_URL.match(url)
        if m:
            return '{base_url}/projects/{project_name}'.format(**m.groupdict())
        else:
            return url

    def get_all_issues(self):
        issues = self.api.get_all_pages(
            '{}/issues.json?status_id=*'.format(self.public_url))
        detailed_issues = []
        # It's impossible to get issue history from list view, so get it from
        # detail view...

        for issue_id in (i['id'] for i in issues):
            issue_url = '{}/issues/{}.json?include=journals,watchers,relations,childrens,attachments'.format(
                self.instance_url, issue_id)
            detailed_issues.append(self.api.get(issue_url))

        return detailed_issues

    def get_participants(self):
        """Get participating users (issues authors/owners)

        :return: list of all users participating on issues
        :rtype: list
        """
        issues = self.get_all_issues()
        return self.get_participants0(issues)

    def get_participants0(self, issues):
        """Get participating users (issues authors/owners)

        :return: list of all users participating on issues
        :rtype: list
        """
        user_ids = set()
        users = []
        for issue in issues:
            for user in chain(issue.get('watchers', []), [issue['author'], issue.get('assigned_to', None)]):
                if user is None:
                    continue
                user_ids.add(user['id'])

        for user_id in user_ids:
            # The anonymous user is not really part of the project...
            if user_id != ANONYMOUS_USER_ID:
                users.append(self.api.get('{}/users/{}.json'.format(self.instance_url, user_id)))
        return users

    def get_users_index(self):
        """ Returns dict index of users (by user id)
        """
        return {i['id']: i for i in self.get_participants()}

    def get_versions(self):
        response = self.api.get('{}/versions.json'.format(self.public_url))
        return response['versions']

    def load_attachment_file(self, attachment):
        return self.api.load(attachment['content_url'])


class RedmineProjectWithCache(Project):
    REGEX_PROJECT_URL = re.compile(
        r'^(?P<base_url>https?://.*)/projects/(?P<project_name>[\w_-]+)$')

    def __init__(self, url, cache_dir, *args, **kwargs):
        # noinspection PyCompatibility
        super().__init__(url, *args, **kwargs)
        self.api_url = '{}.json'.format(self.public_url)
        self.instance_url = self._url_match.group('base_url')
        self.path = cache_dir
        with open(os.path.join(self.path, 'project.json'), 'r') as outfile:
            self.project = json.load(outfile)
        log.info('Got redmine project: {}'.format(self.get_id()))

    def get_id(self):
        return str(self.project['id'])

    def get_project(self):
        return self.project

    def get_all_issues(self):
        data = self._load_data(os.path.join(self.path, 'issues'))
        return sorted(data, key=lambda issue: issue['id'])

    def get_participants(self):
        return self._load_data(os.path.join(self.path, 'users'))

    def get_users_index(self):
        return {i['id']: i for i in self.get_participants()}

    def get_versions(self):
        data = self._load_data(os.path.join(self.path, 'versions'))
        return sorted(data, key=lambda version: version['id'])

    def get_attachments(self):
        return self._load_data(os.path.join(self.path, 'attachments'))

    def get_attachments_index(self):
        return {i['id']: i for i in self.get_attachments()}

    def link_roadmap(self, version, gitlab_id, gitlab_url, cache):
        if not '/milestones/' in version['description']:
            version['description'] += "Moved to {}/milestones/{}".format(gitlab_url, gitlab_id)
        cache.load_version2(version)

    def link_issue(self, issue, gitlab_id, gitlab_url, cache):
        issue['note'] = "Moved to {}/issues/{}".format(gitlab_url, gitlab_id)
        cache.load_issue2(issue)

    @staticmethod
    def _load_data(path):
        result = []
        for entry in os.scandir(path):
            if entry.is_file() and str(entry.path).endswith('.json'):
                with open(entry.path, 'r') as outfile:
                    data = json.load(outfile)
                    result.append(data)
        return result


class RedmineCacheWriter:
    def __init__(self, cache_dir, project):
        self.path = cache_dir
        log.info('Redmine Cache dir: {}'.format(self.path))
        self.project = project
        self._create_dir(self.path)
        self._store_data(self.path, project.get_project(), 'project', 'Project file')

    def load_versions(self):
        path = os.path.join(self.path, 'versions')
        if os.path.exists(path):
            log.info('Load versions from cache {}'.format(path))
            versions = self._load_data(path)
        else:
            self._create_dir(path)
            log.info('Store versions')
            versions = self.project.get_versions()
            for version in versions:
                self._store_data(path, version, version['id'], 'Version')
        return versions

    def load_issues(self):
        path = os.path.join(self.path, 'issues')
        if os.path.exists(path):
            log.info('Load issues from cache {}'.format(path))
            issues = self._load_data(path)
        else:
            self._create_dir(path)
            log.info('Loading issues')
            issues = self.project.get_all_issues()
            for issue in issues:
                self._store_data(path, issue, issue['id'], 'Issue')
        return issues

    def load_users(self, issues):
        path = os.path.join(self.path, 'users')
        if os.path.exists(path):
            log.info('Load users from cache {}'.format(path))
            users = self._load_data(path)
        else:
            self._create_dir(path)
            log.info('Loading users')
            users = self.project.get_participants0(issues)
            for user in users:
                self._store_data(path, user, user['id'], 'User')
        return users

    def load_attachments(self, issues):
        path = os.path.join(self.path, 'attachments')
        if os.path.exists(path):
            log.info('Load attachments from cache {}'.format(path))
            attachments = self._load_data(path)
        else:
            self._create_dir(path)
            log.info('Loading attachments')
            attachments = []
            for issue in issues:
                at = issue['attachments']
                if at:
                    for a in at:
                        attachments.append(a)
                        a_content = self.project.load_attachment_file(a)
                        file = os.path.join(path, '{}.data'.format(a['id']))
                        with open(file, 'wb') as outfile:
                            outfile.write(a_content)
                        a['file'] = file
                        self._store_data(path, a, a['id'], 'Attachment')
        return attachments

    def load_attachment(self, attachment):
        path = os.path.join(self.path, 'attachments')
        self._store_data(path, attachment, attachment['id'], 'Attachment')

    def load_issue(self, issue):
        path = os.path.join(self.path, 'issues')
        self._store_data(path, issue, issue['id'], 'Issue')

    def load_version(self, version):
        path = os.path.join(self.path, 'versions')
        self._store_data(path, version, version['id'], 'Version')

    def load_version2(self, version):
        path = os.path.join(self.path, 'versions2')
        self._create_dir(path)
        v = {
            "version": {
                "name": version['name'],
                "description": version['description']
            }
        }
        self._store_data(path, v, version['id'], 'Version')

    def load_issue2(self, issue):
        path = os.path.join(self.path, 'issues2')
        self._create_dir(path)
        v = {
            "issue": {
                "notes": issue['note']
            }
        }
        self._store_data(path, v, issue['id'], 'Issue')

    def get_version2_path(self, version_id):
        return os.path.join(self.path, 'versions2', '{}.json'.format(version_id))

    def get_issue2_path(self, issue_id):
        return os.path.join(self.path, 'issues2', '{}.json'.format(issue_id))

    @staticmethod
    def _create_dir(path):
        if not os.path.exists(path):
            os.makedirs(path)
            log.info('Create cache dir: {}'.format(path))

    @staticmethod
    def _store_data(path, data, data_id, msg):
        file = os.path.join(path, '{}.json'.format(data_id))
        with open(file, 'w') as outfile:
            json.dump(data, outfile)
        log.info('{} {} to {}'.format(msg, data_id, file))
        log.debug('{} {} = {}'.format(msg, data_id, data))

    @staticmethod
    def _load_data(path):
        result = []
        for entry in os.scandir(path):
            if entry.is_file() and str(entry.path).endswith('.json'):
                with open(entry.path, 'r') as outfile:
                    data = json.load(outfile)
                    result.append(data)
        return result



import logging

import requests

log = logging.getLogger(__name__)


class APIClient:
    def __init__(self, api_key):
        self.api_key = api_key

    def get_auth_headers(self):
        """ Method to be overloaded by child classes

        :return: a dict with auth headers set
        """
        return {}

    def add_auth_headers(self, kwargs):
        _kwargs = kwargs.copy()
        headers = kwargs.get('headers', {})
        headers.update(self.get_auth_headers())
        _kwargs['headers'] = headers
        return _kwargs

    def _req(self, func, *args, **kwargs):
        log.debug('HTTP REQUEST {} {} {}'.format(
            func, args, kwargs))
        kwargs = self.add_auth_headers(kwargs)
        resp = func(*args, **kwargs)
        resp.raise_for_status()
        ret = resp.json()
        log.debug('HTTP RESPONSE {}'.format(ret))
        return ret

    def _req2(self, func, *args, **kwargs):
        log.debug('HTTP REQUEST {} {} {}'.format(
            func, args, kwargs))
        kwargs = self.add_auth_headers(kwargs)
        resp = func(*args, **kwargs)
        resp.raise_for_status()
        ret = resp.content
        log.debug('HTTP RESPONSE {}'.format(len(ret)))
        return ret

    def get(self, *args, **kwargs):
        return self._req(requests.get, *args, **kwargs)

    def post(self, *args, **kwargs):
        return self._req(requests.post, *args, **kwargs)

    def put(self, *args, **kwargs):
        return self._req(requests.put, *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._req(requests.delete, *args, **kwargs)

    def load(self, *args, **kwargs):
        return self._req2(requests.get, *args, **kwargs)


class Project:
    def __init__(self, url, client):
        self.public_url = url.strip('/')  # normalize URL
        self.api = client

        # noinspection PyUnresolvedReferences
        self._url_match = self.REGEX_PROJECT_URL.match(self.public_url)
        if self._url_match is None:
            raise ValueError('{} is not a valid project URL'.format(url))



#!/bin/env python3
import argparse
import logging
import re

import subprocess

from migrate_redmine_to_gitlab import sql
from migrate_redmine_to_gitlab.config import MigrationConfig
from migrate_redmine_to_gitlab.converters import convert_issue, convert_version, convert_attachments
from migrate_redmine_to_gitlab.gitlab import GitlabClient, GitlabProject
from migrate_redmine_to_gitlab.logging import setup_module_logging
from migrate_redmine_to_gitlab.redmine import RedmineClient, RedmineProjectWithCache, RedmineProject, RedmineCacheWriter

"""Migration commands for issues and roadmaps from redmine to gitlab
"""

log = logging.getLogger(__name__)


class CommandError(Exception):
    """ An error that will nicely pop up to user and stops program
    """

    def __init__(self, msg):
        self.msg = msg


def parse_args():
    parser = argparse.ArgumentParser(description=__doc__)
    subparsers = parser.add_subparsers(dest='command')
    commands = []

    init = subparsers.add_parser('init', help=Init.__doc__)
    init.set_defaults(command=Init)
    commands.append(init)

    roadmap = subparsers.add_parser('roadmap', help=Versions.__doc__)
    roadmap.set_defaults(command=Versions)
    commands.append(roadmap)

    attachments = subparsers.add_parser('attachments', help=Attachments.__doc__)
    attachments.set_defaults(command=Attachments)
    commands.append(attachments)

    issues = subparsers.add_parser('issues', help=Issues.__doc__)
    issues.set_defaults(command=Issues)
    commands.append(issues)

    issues_with_id = subparsers.add_parser('issues-with-id', help=IssuesWithId.__doc__)
    issues_with_id.set_defaults(command=IssuesWithId)
    commands.append(issues_with_id)

    delete_issues = subparsers.add_parser('delete-issues', help=DeleteIssues.__doc__)
    delete_issues.set_defaults(command=DeleteIssues)
    commands.append(delete_issues)

    link_roadmap = subparsers.add_parser('link-roadmap', help=LinkRedmineRoadmap.__doc__)
    link_roadmap.set_defaults(command=LinkRedmineRoadmap)
    commands.append(link_roadmap)

    link_issue = subparsers.add_parser('link-issues', help=LinkRedmineIssue.__doc__)
    link_issue.set_defaults(command=LinkRedmineIssue)
    commands.append(link_issue)

    iid = subparsers.add_parser('iid', help=Iid.__doc__)
    iid.set_defaults(command=Iid)
    commands.append(iid)

    for i in commands:
        i.add_argument('--check', required=False, action='store_true', default=False,
                       help="do not perform any action, just check everything is ready")
        i.add_argument('--debug', required=False, action='store_true', default=False, help="More output")

    return parser.parse_args()


def main():
    args = parse_args()
    if args.debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    # Configure global logging
    setup_module_logging('migrate_redmine_to_gitlab', level=log_level)

    config = MigrationConfig()

    try:
        args.command(config, args).run()
    except CommandError as e:
        log.error(e)
        exit(12)


class Command:
    def __init__(self, config, args):
        self.gitlab = None
        self.redmine = None
        self.args = args
        self.config = config
        log.info('Init {}'.format(self))

    def run(self):
        log.info('Run {}'.format(self))
        self.execute()
        log.info('End {}'.format(self))

    def check(self, func, message):
        # noinspection PyCallingNonCallable
        ret = func(redmine=self.redmine, gitlab=self.gitlab)
        if ret:
            log.info('{}... OK'.format(message))
        else:
            log.error('{}... FAILED'.format(message))
            exit(1)

    def redmine_project_with_cache(self):
        redmine_client = RedmineClient(self.config.redmine_key)
        return RedmineProjectWithCache(self.config.redmine_project_url, self.config.cache_dir, redmine_client)

    def redmine_project(self):
        redmine_client = RedmineClient(self.config.redmine_key)
        return RedmineProject(self.config.redmine_project_url, redmine_client)

    def gitlab_project(self):
        gitlab_client = GitlabClient(self.config.gitlab_key)
        return GitlabProject(self.config.gitlab_project_url, gitlab_client)

    def redmine_cache(self, redmine_project):
        return RedmineCacheWriter(self.config.cache_dir, redmine_project)

    def execute(self):
        pass


class Init(Command):
    def __init__(self, config, args):
        # noinspection PyCompatibility
        super().__init__(config, args)
        self.redmine = self.redmine_project()
        self.gitlab = self.gitlab_project()
        self.cache = self.redmine_cache(self.redmine)

    def execute(self):
        versions = self.cache.load_versions()
        log.info('{} version(s) loaded'.format(len(versions)))

        issues = self.cache.load_issues()
        log.info('{} issue(s) loaded'.format(len(issues)))

        users = self.cache.load_users(issues)
        log.info('{} user(s) loaded'.format(len(users)))

        attachments = self.cache.load_attachments(issues)
        log.info('{} attachment(s) loaded'.format(len(attachments)))


class Versions(Command):
    def __init__(self, config, args):
        # noinspection PyCompatibility
        super().__init__(config, args)
        self.redmine = self.redmine_project_with_cache()
        self.gitlab = self.gitlab_project()
        self.cache = self.redmine_cache(self.redmine)

        checks = [
            (self.check_no_milestone, 'Gitlab project has no pre-existing milestone'),
            (self.check_origin_milestone, 'Redmine project contains versions'),
        ]
        for i in checks:
            self.check(*i)

        self.redmine_versions = self.redmine.get_versions()
        log.info('Got {} version(s) from redmine.'.format(len(self.redmine_versions)))

    def execute(self):

        existing_gitlab_versions = self.gitlab.get_milestones_index().keys()

        versions_data = [convert_version(redmine_version) for redmine_version in self.redmine_versions]
        if self.args.check:
            for data, meta in versions_data:
                if not data['title'] in existing_gitlab_versions:
                    log.info("Would create version {}".format(data))
            return

        gitlab_versions = []
        created_versions, bad_versions = self._create_versions(versions_data, [], existing_gitlab_versions)
        gitlab_versions += created_versions
        while len(bad_versions) > 0:
            log.info('Some versions were not created: {}'.format(bad_versions))

            created_versions, bad_versions = self._create_versions(versions_data,
                                                                   bad_versions,
                                                                   existing_gitlab_versions)
            gitlab_versions += created_versions

        versions_index = {str(redmine_version['id']): redmine_version for redmine_version in self.redmine_versions}

        for gitlab_version in gitlab_versions:
            redmine_id = gitlab_version['redmine_id']
            redmine_version = versions_index[redmine_id]
            redmine_version['gitlab_id'] = gitlab_version['id']
            self.cache.load_version(redmine_version)

        log.info('{} version(s) created on GitLab'.format(len(gitlab_versions)))

    # noinspection PyUnusedLocal
    @staticmethod
    def check_origin_milestone(redmine, gitlab):
        return len(redmine.get_versions()) > 0

    # noinspection PyUnusedLocal
    @staticmethod
    def check_no_milestone(redmine, gitlab):
        return len(gitlab.get_milestones()) == 0

    def _create_versions(self, versions_data, incoming_bad_versions, existing_gitlab_versions):
        bad_versions = []
        created_versions = []
        for data, meta in versions_data:
            if data['title'] in existing_gitlab_versions:
                log.info("skip existing milestone {}".format(data['title']))
                pass
            log.debug(data)
            data_id = str(data['redmine_id'])
            if len(incoming_bad_versions) == 0 or data_id in incoming_bad_versions:
                # noinspection PyBroadException
                try:
                    created_version = self.gitlab.create_milestone(data, meta)
                    created_version['redmine_id'] = data_id
                    created_versions.append(created_version)
                    log.info("Version {}".format(created_version['title']))
                except:
                    log.error('Could not create version {}'.format(data_id))
                    bad_versions.append(data_id)
        return created_versions, bad_versions


class Attachments(Command):
    def __init__(self, config, args):
        # noinspection PyCompatibility
        super().__init__(config, args)
        self.redmine = self.redmine_project_with_cache()
        self.gitlab = self.gitlab_project()
        self.cache = self.redmine_cache(self.redmine)

        self.redmine_issues = self.redmine.get_all_issues()
        log.info('Got {} issue(s) from redmine.'.format(len(self.redmine_issues)))

        self.attachments_index = self.redmine.get_attachments_index()
        log.info('Got {} attachment(s) from redmine.'.format(len(self.attachments_index.values())))

    def execute(self):

        gitlab_id = self.gitlab.get_id()

        attachments_data = []

        for redmine_issue in self.redmine_issues:
            attachments_data += convert_attachments(redmine_issue, self.attachments_index, gitlab_id)

        if self.args.check:
            for data in attachments_data:
                log.info('Would create attachment "{}"'.format(data['redmine']['filename']))
            return

        gitlab_attachments = []
        created_attachments, bad_attachments = self._create_attachments(attachments_data, [])
        gitlab_attachments += created_attachments
        while len(bad_attachments) > 0:
            log.info('Some attachments were not created: {}'.format(bad_attachments))

            created_attachments, bad_attachments = self._create_attachments(attachments_data, bad_attachments)
            gitlab_attachments += created_attachments

        for attachment in gitlab_attachments:
            redmine_attachment = attachment['redmine']
            redmine_attachment['gitlab'] = attachment['gitlab']
            self.cache.load_attachment(redmine_attachment)

        log.info('{} attachments(s) created on GitLab'.format(len(gitlab_attachments)))

    def _create_attachments(self, attachments_data, incoming_bad_attachments):
        bad_attachments = []
        created_attachments = []
        for data in attachments_data:
            data_redmine_ = data['redmine']
            data_id = str(data_redmine_['id'])
            if len(incoming_bad_attachments) == 0 or data_id in incoming_bad_attachments:
                # noinspection PyBroadException
                try:
                    created_attachment = self.gitlab.create_attachment(data)
                    data['gitlab'] = created_attachment
                    created_attachments.append(data)
                    log.info("Created attachment (was: {}) {}".format(data_id, created_attachment['markdown']))
                except:
                    bad_attachments.append(data_id)
                    # noinspection SpellCheckingInspection
                    log.error(
                        'Could not create attachment {} {} (size: {})'.format(data_id, data_redmine_['filename'],
                                                                              data_redmine_['filesize']))
        return created_attachments, bad_attachments


class Issues(Command):
    def __init__(self, config, args):
        # noinspection PyCompatibility
        super().__init__(config, args)
        self.redmine = self.redmine_project_with_cache()
        self.gitlab = self.gitlab_project()
        self.cache = self.redmine_cache(self.redmine)

        self.gitlab_users_index = self.gitlab.get_users_index()
        log.info('Got {} users(s) from gitlab.'.format(len(self.gitlab_users_index.values())))

        self.redmine_users_index = self.redmine.get_users_index()
        log.info('Got {} users(s) from redmine.'.format(len(self.redmine_users_index.values())))

        checks = [
            (self.check_users, 'Required users presence'),
            (self.check_no_issue, 'Project has no pre-existing issue'),
        ]
        for i in checks:
            self.check(*i)

        self.redmine_issues = self.redmine.get_all_issues()
        log.info('Got {} issue(s) from redmine.'.format(len(self.redmine_issues)))

        self.attachments_index = self.redmine.get_attachments_index()
        log.info('Got {} attachment(s) from redmine.'.format(len(self.attachments_index.values())))

        self.milestones_index = self.gitlab.get_milestones_index()
        log.info('Got {} milestone(s) from gitlab.'.format(len(self.milestones_index.values())))

        self.gitlab_id = self.gitlab.get_id()

    def execute(self):

        issues_data = self._convert_issues()

        if self.args.check:
            for data, meta in issues_data:
                milestone_id = data.get('milestone_id', None)
                if milestone_id:
                    try:
                        self.gitlab.get_milestone_by_id(milestone_id)
                    except ValueError:
                        raise CommandError(
                            "issue \"{}\" points to unknown milestone_id \"{}\". "
                            "Check that you already migrated roadmaps".format(
                                data['title'], milestone_id))

                log.info(
                    'Would create issue "{}" with {} notes and {} attachments.'.format(data['title'],
                                                                                       len(meta['notes']),
                                                                                       len(meta['attachments'])))
            return

        gitlab_issues = []

        created_issues, bad_issues = self._create_issues(issues_data, [])
        gitlab_issues += created_issues
        while len(bad_issues) > 0:
            log.info('Some issues were not created: {}'.format(bad_issues))
            created_issues, bad_issues = self._create_issues(issues_data, bad_issues)
            gitlab_issues += created_issues

        issues_by_index = {str(redmine_issue['id']): redmine_issue for redmine_issue in self.redmine_issues}

        for gitlab_issue in gitlab_issues:
            redmine_id = gitlab_issue['redmine_id']
            redmine_issue = issues_by_index[redmine_id]
            redmine_issue['gitlab_id'] = gitlab_issue['iid']
            self.cache.load_issue(redmine_issue)
        log.info('{} issue(s) created on GitLab'.format(len(gitlab_issues)))

    def _convert_issues(self):
        return [
            convert_issue(redmine_issue,
                          self.redmine_users_index,
                          self.attachments_index,
                          self.gitlab_id,
                          self.gitlab_users_index,
                          self.milestones_index) for redmine_issue in self.redmine_issues]

    # noinspection PyUnusedLocal
    @staticmethod
    def check_no_issue(redmine, gitlab):
        return len(gitlab.get_issues()) == 0

    @staticmethod
    def check_users(redmine, gitlab):
        users = redmine.get_participants()
        # Filter out anonymous user
        nicks = set([i['login'] for i in users if i['login'] != ''])
        log.info('Project users are: {}'.format(', '.join(nicks) + ' '))

        gitlab_user_names = set([i['username'] for i in gitlab.get_all_users()])
        return all((i in gitlab_user_names for i in nicks))

    def _create_issues(self, issues_data, incoming_bad_issues):
        bad_issues = []
        created_issues = []
        for data, meta in issues_data:
            data_id = str(data['redmine_id'])
            if len(incoming_bad_issues) == 0 or data_id in incoming_bad_issues:
                # noinspection PyBroadException
                try:
                    created_issue = self.gitlab.create_issue(data, meta)
                    created_issue['redmine_id'] = data_id
                    created_issues.append(created_issue)
                    log.info("Created issue (was: {}) {}".format(data_id, created_issue['title']))
                except:
                    log.error('Could not create issue {}'.format(data_id))
                    bad_issues.append(data_id)
        return created_issues, bad_issues


class IssuesWithId(Issues):
    def __init__(self, config, args):
        # noinspection PyCompatibility
        super().__init__(config, args)

    def _convert_issues(self):
        return [
            convert_issue(redmine_issue,
                          self.redmine_users_index,
                          self.attachments_index,
                          self.gitlab_id,
                          self.gitlab_users_index,
                          self.milestones_index,
                          with_id=True) for redmine_issue in self.redmine_issues]


class Iid(Command):
    def __init__(self, config, args):
        # noinspection PyCompatibility
        super().__init__(config, args)
        self.redmine = self.redmine_project_with_cache()
        self.gitlab = self.gitlab_project()
        self.cache = self.redmine_cache(self.redmine)

    def execute(self):

        gitlab_project_id = self.gitlab.get_id()

        regex_saved_iid = r'-RM-([0-9]+)-MR-(.*)'

        sql_cmd = sql.COUNT_UNMIGRATED_ISSUES.format(
            regex=regex_saved_iid, project_id=gitlab_project_id)

        output = sql.run_query(sql_cmd)

        try:
            m = re.match(r'\s*(\d+)\s*', output, re.DOTALL | re.MULTILINE)
            issues_count = int(m.group(1))
        except (AttributeError, ValueError):
            raise ValueError(
                'Invalid output from postgres command: "{}"'.format(output))

        if issues_count > 0:
            log.info('Ready to recover iid for {} issues.'.format(
                issues_count))
        else:
            log.error(
                "No issue to migrate iid, possible causes: "
                "you already migrated iid or you haven't migrated issues yet.")
            exit(1)

        if not self.args.check:
            sql_cmd = sql.MIGRATE_IID_ISSUES.format(
                regex=regex_saved_iid, project_id=gitlab_project_id)
            out = sql.run_query(sql_cmd)
            log.info(out)
            try:
                m = re.match(
                    r'\s*(\d+)\s*', output,
                    re.DOTALL | re.MULTILINE)
                migrated_count = int(m.group(1))
                log.info('Migrated successfully iid for {} issues'.format(
                    migrated_count))
            except (IndexError, AttributeError):
                raise ValueError(
                    'Invalid output from postgres command: "{}"'.format(output))


class DeleteIssues(Command):
    def __init__(self, config, args):
        # noinspection PyCompatibility
        super().__init__(config, args)
        self.gitlab = self.gitlab_project()

    def execute(self):
        log.info('Start {}'.format(self))

        gitlab_issues = self.gitlab.get_issues()
        log.info('Got {} issue(s) from gitlab.'.format(len(gitlab_issues)))

        for issue in gitlab_issues:
            log.info('delete issue {}'.format(issue['id']))
            self.gitlab.delete_issue(issue['id'])


class LinkRedmineRoadmap(Command):
    def __init__(self, config, args):
        # noinspection PyCompatibility
        super().__init__(config, args)
        self.redmine = self.redmine_project_with_cache()
        self.gitlab = self.gitlab_project()
        self.cache = self.redmine_cache(self.redmine)

    def execute(self):
        log.info('Start {}'.format(self))

        redmine_versions = self.redmine.get_versions()
        log.info('Got {} version(s) from redmine.'.format(len(redmine_versions)))

        gitlab_milestones = self.gitlab.get_milestones()
        log.info('Got {} milestone(s) from gitlab.'.format(len(gitlab_milestones)))
        gitlab_milestones_index = {i['title']: i for i in gitlab_milestones}

        for redmine_version in redmine_versions:

            name = redmine_version['name']
            gitlab_milestone = gitlab_milestones_index[name]
            log.debug(gitlab_milestone)
            gitlab_milestone_iid = gitlab_milestone['iid']
            self.redmine.link_roadmap(redmine_version, gitlab_milestone_iid, self.gitlab.project_url, self.cache)

            redmine_version_id = redmine_version['id']
            json_file_path = self.cache.get_version2_path(redmine_version_id)

            action = ['curl', '-v', '-X', 'PUT',
                      '-H', 'Content-Type: application/json',
                      '-H', 'X-Redmine-API-Key:{}'.format(self.config.redmine_key),
                      '--data-binary', '@{}'.format(json_file_path),
                      '{}/versions/{}.json'.format(self.config.redmine_host, redmine_version_id)]
            if self.args.check:
                log.info(
                    'Would link redmine version {} ({}) to gitlab milestone {}'.format(redmine_version_id, name,
                                                                                       gitlab_milestone_iid))
                log.info('using command: {}'.format(action))
            else:
                log.info('Link redmine version {} ({}) to gitlab milestone {}'.format(redmine_version_id, name,
                                                                                      gitlab_milestone_iid))
                log.debug('using command: {}'.format(action))
                subprocess.run(action, stdout=subprocess.PIPE)


class LinkRedmineIssue(Command):
    def __init__(self, config, args):
        # noinspection PyCompatibility
        super().__init__(config, args)
        self.redmine = self.redmine_project_with_cache()
        self.gitlab = self.gitlab_project()
        self.cache = self.redmine_cache(self.redmine)

    def execute(self):
        log.info('Start {}'.format(self))

        redmine_issues = self.redmine.get_all_issues()

        log.info('Got {} issue(s) from redmine.'.format(len(redmine_issues)))

        gitlab_issues = self.gitlab.get_issues()
        gitlab_issues_index = {i['title']: i for i in gitlab_issues}

        log.info('Got {} issue(s) from gitlab.'.format(len(gitlab_issues)))

        for redmine_issue in redmine_issues:
            subject = redmine_issue['subject'].strip()
            gitlab_issue = gitlab_issues_index[subject]
            redmine_issue_id = redmine_issue['id']
            self.redmine.link_issue(redmine_issue, gitlab_issue['iid'], self.gitlab.project_url, self.cache)

            json_file_path = self.cache.get_issue2_path(redmine_issue_id)

            action = ['curl', '-X', 'PUT',
                      '-H', 'Content-Type: application/json',
                      '-H', 'X-Redmine-API-Key:{}'.format(self.config.redmine_key),
                      '--data-binary', '@{}'.format(json_file_path),
                      '{}/issues/{}.json'.format(self.config.redmine_host, redmine_issue_id)]
            if self.args.check:
                log.info(
                    'Would link redmine issue {} ({}) to {}'.format(redmine_issue_id, subject, gitlab_issue['iid']))
                log.info('using command: {}'.format(action))
            else:
                log.info('Link redmine issue {} ({}) to {}'.format(redmine_issue_id, subject, gitlab_issue['iid']))
                log.debug('using command: {}'.format(action))
                subprocess.run(action, shell=True, check=True)



import logging
import subprocess

""" SQL-related work for gitlab DB
"""

log = logging.getLogger(__name__)


COUNT_UNMIGRATED_ISSUES = r"""
SELECT COUNT(*)
FROM issues
WHERE title ~* '{regex}' AND project_id={project_id};
"""


MIGRATE_IID_ISSUES = r"""
UPDATE issues SET
  title = regexp_replace(issues.title, '{regex}','\2'),
  iid = regexp_replace(issues.title, '{regex}', '\1')::integer
WHERE title ~* '{regex}' AND project_id={project_id};
"""


def run_query(
        cmd,
        unix_user='gitlab-psql',
        hostname='/var/opt/gitlab/postgresql',
        dbname='gitlabhq_production',
        psql_bin='/opt/gitlab/embedded/bin/psql'):
    """Run a sql command and returns output

    Defaults match omnibus-installed gitlab settings.

    :param cmd: a SQL command, ending with ";"
    :type cmd: str
    :rtype: str
    """

    log.debug('Running SQL command {}'.format(cmd))

    output = subprocess.check_output([
        'sudo', '-u', unix_user,
        psql_bin,
        '-A', '-t',  # supress output fancy
        '-h', hostname,
        '-d', dbname,
    ], input=cmd.encode())
    log.debug('SQL output is "{}"'.format(output.decode()))

    return output.decode()



import logging

log = logging.getLogger(__name__)


def redmine_uid_to_login(redmine_id, redmine_user_index):
    return redmine_user_index[redmine_id]['login']


def redmine_uid_to_gitlab_uid(redmine_id,
                              redmine_user_index, gitlab_user_index):
    username = redmine_uid_to_login(redmine_id, redmine_user_index)
    return gitlab_user_index[username]['id']


def convert_notes(redmine_issue_journals, redmine_user_index):
    """ Convert a list of redmine journal entries to gitlab notes

    Filters out the empty notes (ex: bare status change)
    Adds metadata as comment

    :param redmine_issue_journals: list of redmine "journals"
    :param redmine_user_index: dictionary of redmine users
    :return: yielded couple ``data``, ``meta``. ``data`` is the API payload for
        an issue note and meta a dict (containing, at the moment, only a "sudo_user" key).
    """

    for entry in redmine_issue_journals:
        journal_notes = entry.get('notes', '')
        if not 'Migrated to https' in journal_notes \
                and not 'Moved to https' in journal_notes \
                and len(journal_notes) > 0:
            body = "{}\n\n*(from redmine: written on {})*".format(
                journal_notes, entry['created_on'][:10])
            try:
                author = redmine_uid_to_login(
                    entry['user']['id'], redmine_user_index)
            except KeyError:
                # In some cases you have anonymous notes, which do not exist in
                # gitlab.
                log.debug(
                    'Redmine user {} is unknown, attribute note '
                    'to current admin\n'.format(entry['user']))
                author = None
            yield {'body': body}, {'sudo_user': author}


def relations_to_string(relations, issue_id):
    """ Convert redmine formal relations to some un-normalized string

    That's the way gitlab does relations, by "mentioning".

    :param relations: list of issues relations
    :param issue_id: the current issue id
    :return a string listing relations.
    """
    l = []
    for i in relations:
        if issue_id == i['issue_id']:
            other_issue_id = i['issue_to_id']
        else:
            other_issue_id = i['issue_id']
        l.append('{} #{}'.format(i['relation_type'], other_issue_id))

    return ', '.join(l)


def convert_issue(redmine_issue,
                  redmine_user_index,
                  redmine_attachments_index,
                  gitlab_project_id,
                  gitlab_user_index,
                  gitlab_milestones_index,
                  with_id=False):
    if redmine_issue.get('closed_on', None):
        # quick'n dirty extract date
        close_text = ', closed on {}'.format(redmine_issue['closed_on'][:10])
        closed = True
    else:
        close_text = ''
        closed = False

    relations = redmine_issue.get('relations', [])
    relations_text = relations_to_string(relations, redmine_issue['id'])
    if len(relations_text) > 0:
        relations_text = ', ' + relations_text

    if with_id:
        title = '-RM-{}-MR-{}'.format(redmine_issue['id'], redmine_issue['subject'])
    else:
        title = redmine_issue['subject']

    labels = 'From Redmine, ' + redmine_issue['tracker']['name'] + ', ' + redmine_issue['priority']['name']

    data = {
        'title': title,
        'redmine_id': redmine_issue['id'],
        'description': '{}\n\n*(from redmine issue {} created on {}{}{})*'.format(
            redmine_issue['description'],
            redmine_issue['id'],
            redmine_issue['created_on'][:10],
            close_text,
            relations_text
        ),
        'labels': labels
    }

    version = redmine_issue.get('fixed_version', None)
    if version:
        data['milestone_id'] = gitlab_milestones_index[version['name']]['id']

    try:
        author_login = redmine_uid_to_login(
            redmine_issue['author']['id'], redmine_user_index)

    except KeyError:
        log.debug(
            'Redmine issue #{} is anonymous, gitlab issue is attributed '
            'to current admin\n'.format(redmine_issue['id']))
        author_login = None

    attachments = []
    for a in redmine_issue.get('attachments', []):
        attachment = redmine_attachments_index[a['id']]
        attachments.append(convert_attachment(gitlab_project_id, attachment))

    meta = {
        'sudo_user': author_login,
        'notes': list(convert_notes(redmine_issue['journals'], redmine_user_index)),
        'must_close': closed,
        'attachments': attachments
    }

    assigned_to = redmine_issue.get('assigned_to', None)
    if assigned_to is not None:
        data['assignee_id'] = redmine_uid_to_gitlab_uid(assigned_to['id'], redmine_user_index, gitlab_user_index)
    return data, meta


def convert_version(redmine_version):
    """ Turns a redmine version into a gitlab milestone

    Do not handle the issues linked to the milestone/version.
    Note that redmine do not expose a due date in API.

    :param redmine_version: a dict describing redmine-api-style version
    :rtype: couple: dict, dict
    :return: a dict describing gitlab-api-style milestone and another for meta
    """
    milestone = {
        "title": redmine_version['name'],
        'redmine_id': redmine_version['id'],
        "description": '{}\n\n*(from redmine: created on {})*'.format(
            redmine_version['description'],
            redmine_version['created_on'][:10])
    }
    if 'due_date' in redmine_version:
        milestone['due_date'] = redmine_version['due_date'][:10]

    must_close = redmine_version['status'] == 'closed'

    return milestone, {'must_close': must_close}


def convert_attachments(redmine_issue, redmine_attachments_index, gitlab_project_id):
    attachments = []
    for redmine_attachment in redmine_issue.get('attachments', []):
        attachment = redmine_attachments_index[redmine_attachment['id']]
        attachments.append(convert_attachment(gitlab_project_id, attachment))

    return attachments


def convert_attachment(gitlab_project_id, redmine_attachment):
    return {"redmine": redmine_attachment, "request": {"id": gitlab_project_id, "file": redmine_attachment['filename']}}



import re
import logging
from . import APIClient, Project

log = logging.getLogger(__name__)


class GitlabClient(APIClient):
    # see http://doc.gitlab.com/ce/api/#pagination
    MAX_PER_PAGE = 100

    def get(self, *args, **kwargs):
        # Note that we do not handle pagination, but as we rely on list data
        # only for milestones, we assume that we have < 100 milestones. Could
        # be fixed though...
        kwargs['params'] = kwargs.get('params', {})
        kwargs['params']['per_page'] = self.MAX_PER_PAGE
        # noinspection PyCompatibility
        return super().get(*args, **kwargs)

    def get_auth_headers(self):
        return {"PRIVATE-TOKEN": self.api_key}

    def check_is_admin(self):
        pass


class GitlabProject(Project):
    REGEX_PROJECT_URL = re.compile(
        r'^(?P<base_url>https?://.*/)(?P<namespace>[^/]+)/(?P<project_name>[\w_-]+)$')

    def __init__(self, *args, **kwargs):
        # noinspection PyCompatibility
        super().__init__(*args, **kwargs)
        self.api_url0 = ('{base_url}api/v3/projects/{namespace}%2F{project_name}'.format(**self._url_match.groupdict()))
        self.project = self.get_project()
        self.project_id = str(self.project['id'])
        self.project_url = '{base_url}{namespace}/{project_name}'.format(**self._url_match.groupdict())
        log.info('Go gitlab project {}'.format(self.project_id))
        self.api_url = (('{base_url}api/v3/projects/' + self.project_id).format(**self._url_match.groupdict()))
        self.instance_url = '{}/api/v3'.format(self._url_match.group('base_url'))

    def is_repository_empty(self):
        """ Heuristic to check if repository is empty
        """
        return self.api.get(self.api_url)['default_branch'] is None

    def create_attachment(self, data):
        """ High-level attachment creation

        :param data: dict formatted as the gitlab API expects it
        :return: the created attachment
        """
        attachment_url = '{}/uploads'.format(self.api_url)
        data_redmine_ = data['redmine']
        content_type = data_redmine_.get('content_type', 'application/text')
        files = {'file': (data_redmine_['filename'], open(data_redmine_['file'], 'rb'), content_type)}

        return self.api.post(attachment_url, data['request'], files=files)

    def create_issue(self, data, meta):
        """ High-level issue creation

        :param meta: dict with "sudo_user", "should_close", "attachments" and "notes" keys
        :param data: dict formatted as the gitlab API expects it
        :return: the created issue (without notes)
        """
        if len(meta['attachments']) > 0:
            attachment_log = '\n\n### Files'
            for attachment in meta['attachments']:
                redmine_attachment = attachment['redmine']
                gitlab_attachment = redmine_attachment.get('gitlab', None)
                if gitlab_attachment is None:
                    attachment_log += '\n  * [{}] ({})'.format(redmine_attachment['filename'],
                                                               redmine_attachment['content_url'])
                else:
                    attachment_log += '\n  * [{}] ({})'.format(gitlab_attachment['alt'], gitlab_attachment['url'])
            data['description'] += attachment_log

        issues_url = '{}/issues'.format(self.api_url)
        issue = self.api.post(issues_url, data=data)

        issue_url = '{}/{}'.format(issues_url, issue['id'])

        # Handle issues notes
        issue_notes_url = '{}/notes'.format(issue_url, 'notes')
        for note_data, note_meta in meta['notes']:
            self.api.post(issue_notes_url, data=note_data)

        # Handle closed status
        if meta['must_close']:
            altered_issue = issue.copy()
            altered_issue['labels'] = data['labels']
            altered_issue['state_event'] = 'close'
            self.api.put(issue_url, data=altered_issue)

        return issue

    def create_milestone(self, data, meta):
        """ High-level milestone creation

        :param meta: dict with "should_close"
        :param data: dict formatted as the gitlab API expects it
        :return: the created milestone
        """
        milestones_url = '{}/milestones'.format(self.api_url)
        milestone = self.api.post(milestones_url, data=data)

        if meta['must_close']:
            milestone_url = '{}/{}'.format(milestones_url, milestone['id'])
            altered_milestone = milestone.copy()
            altered_milestone['state_event'] = 'close'

            self.api.put(milestone_url, data=altered_milestone)
        return milestone

    def delete_issue(self, issue_id):
        self.api.delete('{}/issues/{}'.format(self.api_url, issue_id))

    def get_issues(self):
        issues = []
        page = 1
        found_issues = self.api.get('{}/issues?page={}&per_page=100'.format(self.api_url, page))
        while len(found_issues) > 0:
            issues += found_issues
            page += 1
            found_issues = self.api.get('{}/issues?page={}&per_page=100'.format(self.api_url, page))
        return issues

    def get_members(self):
        return self.api.get('{}/members'.format(self.api_url))

    def get_milestones(self):
        if not hasattr(self, '_cache_milestones'):
            # noinspection PyAttributeOutsideInit
            self._cache_milestones = []
            page = 1
            found_issues = self.api.get('{}/milestones?page={}&per_page=100'.format(self.api_url, page))
            while len(found_issues) > 0:
                self._cache_milestones += found_issues
                page += 1
                found_issues = self.api.get('{}/milestones?page={}&per_page=100'.format(self.api_url, page))
        return self._cache_milestones

    def get_milestones_index(self):
        return {i['title']: i for i in self.get_milestones()}

    def get_milestone_by_id(self, _id):
        milestones = self.get_milestones()
        for i in milestones:
            if i['id'] == _id:
                return i
        raise ValueError('Could not get milestone')

    # noinspection SpellCheckingInspection
    def has_members(self, usernames):
        gitlab_user_names = set([i['username'] for i in self.get_members()])
        return all((i in gitlab_user_names for i in usernames))

    def get_project(self):
        return self.api.get(self.api_url0)

    def get_id(self):
        return self.project_id

    def get_all_users(self):
        return self.api.get('{}/users'.format(self.api_url))

    def get_users_index(self):
        """ Returns dict index of users (by login)
        """
        return {i['username']: i for i in self.get_all_users()}



#!/usr/bin/env python3
from __future__ import unicode_literals

import os

from setuptools import setup

try:
    README = open(os.path.join(os.path.dirname(__file__), 'README.md')).read()
except IOError:
    README = ''

setup(
    name='migrate-redmine-to-gitlab',
    version='2.0.2',
    description='Migrate a redmine project to Gitlab',
    long_description=README,
    author='Tony Chemit',
    author_email='dev@tchemit.fr',
    license='GPL',
    url='https://github/ultreia-io/migrate-redmine-to-gitlab',
    packages=['migrate_redmine_to_gitlab'],
    install_requires=['requests'],
    entry_points={
        'console_scripts': [
            'migrate-redmine-to-gitlab = migrate_redmine_to_gitlab.commands:main'
        ]
    },
    test_suite='migrate_redmine_to_gitlab.tests',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
    ]
)
